# PAPI CertAuth functions
# Copyright (c) 2001-2005, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.
#
# This module implements user authentication using a X509 Certificate
# provided by users through their browsers
#

package PAPI::CertAuth;
use strict;
use Time::Local;

sub ValidateCert {
   my $con = shift;
   my $cfg = shift;
   my %months = (Jan => 0, Feb =>1, Mar => 2, Apr => 3, May => 4, Jun => 5, Jul =>6, Aug => 7, Sep => 8, Oct => 9, Nov => 10, Dec => 11);

   if (!exists $$cfg{CompatEnvVars}){
	$$con{PAPIerror}="CompatEnvVars not defined in configuration";
	return undef;
   }
   # Check for  environment variables passed by the SSL module
   # (stated in httpd.conf)
   if (!exists $ENV{SSL_CLIENT_CERTIFICATE}){
	$$con{PAPIerror}="Error in httpd.conf: SSLOptions must include +CompatEnvVars";
	return undef;
   }

   if (!exists $$cfg{CertValidateMethod} or $$cfg{CertValidateMethod} eq ""){
	$$con{PAPIerror}="CertValidateMethod not defined in configuration";
	return undef;
   }
   # Go through the configured cerification methods
   my @mth = split /\s/ , $$cfg{CertValidateMethod};
   foreach (@mth){
	# Look for the corresponding verification pattern
	if (!exists $$cfg{$_."Pattern"} or $$cfg{$_."Pattern"} eq ""){
		$$con{PAPIerror}=$_."Pattern not defined in configuration";
		return undef;
	}
	else{
		# Remote IP address
		if ($_ eq "remote_addr"){
			my @ips= split /\s/, $$cfg{remote_addrPattern};
			my $ok=0;
			foreach my $i (@ips){
				# Address range
				if (($i=~ /(\d+\.\d+\.\d+\.\d+)\/(\d+\.\d+\.\d+\.\d+)/) or ($i=~ /\d+/)){

					$ok=MatchIP($ENV{REMOTE_ADDR},$i);
					last if ($ok);
				}
				# Single address
				elsif ($ENV{REMOTE_ADDR} eq $i){
					$ok=1;
					last;
				}
			}
			if (!$ok){
				$$con{PAPIerror}="Invalid client certificate according to remote_addr pattern";
                        	return undef;
			}
  		}
		# Cert issued after a certain date
		elsif ($_ eq "start_date"){
			my @aux=split /\s/, $ENV{SSL_CLIENT_CERT_START};
			my @hour=split /:/,$aux[2];
			my @date=($hour[2],$hour[1],$hour[0],$aux[1],$months{$aux[0]},$aux[3]-1900);
		
			@aux=split /\s/, $$cfg{start_datePattern};
			@hour=split /:/,$aux[2];
			my @date_template=($hour[2],$hour[1],$hour[0],$aux[1],$months{$aux[0]},$aux[3]-1900);
	 		if (!(&timegm(@date) ge &timegm(@date_template))){ 
				$$con{PAPIerror}="Invalid client certificate according to start_date pattern";
				return undef;
			}       	
		}
		# Cert issued before a certain date
        	elsif ($_ eq "end_date"){
                	my @aux=split /\s/, $ENV{SSL_CLIENT_CERT_START};
                	my @hour=split /:/,$aux[2];
                	my @date=($hour[2],$hour[1],$hour[0],$aux[1],$months{$aux[0]},$aux[3]-1900);
	                @aux=split /\s/, $$cfg{end_datePattern};
        	        @hour=split /:/,$aux[2];
                	my @date_template=($hour[2],$hour[1],$hour[0],$aux[1],$months{$aux[0]},$aux[3]-1900);
	                if (!(&timegm(@date) le &timegm(@date_template))){
				$$con{PAPIerror}="Invalid client certificate according to end_date pattern";
	                        return undef;
        	        }
	        }
		# Subject DN
		elsif ($_ eq "client_DN"){
			if ( !($ENV{SSL_CLIENT_DN}=~ m/$$cfg{client_DNPattern}/gi)){
				$$con{PAPIerror}="Invalid client certificate according to client_DN pattern";
				return undef;
			}	
		}
		# Issuer DN
		elsif ($_ eq "issuer_DN"){
			if ( !($ENV{SSL_CLIENT_IDN}=~ m/$$cfg{issuer_DNPattern}/gi)){
				$$con{PAPIerror}="Invalid client certificate according to issuer_DN pattern";
                                return undef;
                        }
		}
		else {
			$$con{PAPIerror}="Unknown verification method";
			return undef;
		}
	}
   }  
# If a specific attribute has been selected to be returned, read it
# Otherwise, return the whole DN
   my $rt = $ENV{SSL_CLIENT_DN};
   if (exists $$cfg{IdAttribute} && $$cfg{IdAttribute} ne ""){
	if ($ENV{SSL_CLIENT_DN}=~ /(.*)($$cfg{IdAttribute}=.*)(.*)/){
		my @uid=split /=/, $2;
		$rt = $uid[1];
   	}
   }
   if (!defined $rt) { $$con{PAPIerror}="DN or IdAttribute not found in certificate"; }
   return $rt;
}

# This auxiliary function checks whether an IP address is within a range
sub MatchIP{
    my $remip=shift;
    my $range=shift;
    my ($ip,$mask,@vmask,@vip,$bmask,$bip,$bredi,$yuyu,@bredo,$ip_bin);

    ($ip,$mask)=split(/\//,$range);
    @vip=split(/\./,$ip);

    if ($mask=~ /(\d+\.\d+\.\d+\.\d+)/){
	@vmask=split(/\./,$mask);
    	$bmask=pack "C4",$vmask[0],$vmask[1],$vmask[2],$vmask[3];
    }
    # Classless
    elsif ($mask=~ /\d+/){
	$bmask=pack "B32", '1' x $mask . '0' x (32-$mask);
    }
    $bmask= ~$bmask;

    $bip=pack "C4",$vip[0],$vip[1],$vip[2],$vip[3];
    $bip=$bip | $bmask;

    @bredo=split(/\./,$remip);
    $ip_bin=pack "C4",$bredo[0],$bredo[1],$bredo[2],$bredo[3];

    $bredi=$ip_bin | $bmask;
    $bredi=$bip ^ $bredi;
    $yuyu=unpack("I",$bredi);

    return 1 if (!$yuyu);
    return 0;
}

1;
