# PAPI POPAuth functions
# Copyright (c) 2001-2005, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.
#
# These functions implement PAPI user authentication using the POP3 protocol.
#

package PAPI::POPAuth;
use strict;
use Mail::POP3Client;

# This function is intended to be hooked at the authenticationHook in a
# PAPI AuthServer. It uses the connection variables 'username' and
# 'password' to validate them against the POP3 server defined by the
# configuration variable 'pop3Server'.
# The method used for the validation process depends on the value of the
# variable 'pop3Method': 
#    PASS =>     Use normal POP3 login (send USER and PASS commands). This is
#                the default
#    APOP =>     Use the APOP procedure
#    CRAM-MD5 => Use CRAM-MD5 when sending the password
#    BEST =>     Query server capabilities and select the best available
#                method (preference order is APOP, CRAM-MD5, and PASS)
# It is possible to select the use of SSL by setting the variable
# 'pop3SSL' to 1.
# The user id returned to the AuthServer depends on the configuration
# variable 'pop3UID', that may take the values:
#    POPUSER    => The format is "username", this is the default
#    POPSERVER  => The format is "username@pop3Server"
#
sub POP3User {
   my $con = shift;
   my $cfg = shift;
   my $popsrv = "pop";
   my $meth = "PASS";
   my $tls = 0;
   my $popl = undef;

   $popsrv = $$cfg{'pop3Server'} if (exists $$cfg{'pop3Server'});
   $meth = $$cfg{'pop3Method'} if (exists $$cfg{'pop3Method'});
   $tls = $$cfg{'pop3SSL'} if (exists $$cfg{'pop3SSL'});
   my $pop = Mail::POP3Client->new (USER => $$con{'username'},
                                    PASSWORD => $$con{'password'},
                                    HOST => $popsrv,
                                    AUTH_MODE => $meth,
                                    USESSL => $tls);
   if ($pop->Count() >= 0) {
      my $rt = $$con{'username'};
      if (exists $$cfg{'pop3UID'} && $$cfg{'pop3UID'} eq "POPSERVER") {
         $rt .= "\@$popsrv";
      }
      return $rt;
   }
   else {
      $$con{'PAPIerror'} = "POP login failure";
      return undef;
   }
}

1;
