# Copyright (c) 2001-2005, RedIRIS. All Rights Reserved.
## Parts of the code in this file have been contributed by
# Roman Medina, Telefonica Moviles.
## Parts of the code in this file have been contributed by 
#  Luis Melendez, University of Cordoba.
## Parts of the code in this file have been contributed by 
#  Maja Gorecka-Wolniewicz, Nicolaus Copernicus University, Torun, Poland.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.

package PAPI::RewritingProxy;

use strict;
use PAPI::ApachePoA;

#
## Changes to port PAPI to Apache2 and mod_perl2
#
##  use Apache::Constants qw(:common);
#
use Apache2::Const qw(:common);
use Apache2::RequestRec ();
use Apache2::RequestUtil ();
use Apache2::Connection();
use Apache2::ServerRec ();
use Apache2::Util qw(:all);
use APR::URI ();
use Apache2::URI ();

#
## End of the changes
#
use Encode::Guess;

use LWP::UserAgent;
use URI::URL;
use HTML::TokeParser;
use HTML::TreeBuilder;
use HTTP::Cookies;
use HTTP::Request::Common;
use HTTP::Request::Form;
use Data::Dumper;
use CGI;
use vars '$paquete';
use vars '$acum';

# Variables available for proxy specification
use vars '$papx_hostname';
use vars '$papx_hostdest';
use vars '$papx_pathdest';
use vars '$papx_namedest';
use vars '$papx_location';
use vars '$papx_fulldest';
# Backwards compatibility with older proxy specs
use vars '$host_dest';
use vars '$path_dest';
use vars '$name_dest';
use vars qw($VERSION @ISA @EXPORT);
#Modificacion para los log.
use vars RPLF;
use constant RPLFName => '/tmp/RewritingProxyLog';
$|=1;

# Danis changes:
## This variable is unnecessary, only is used to build the string for log messages.
## besides the variable is not used in debug mode.
use vars '$paquete';
$paquete = __PACKAGE__;

# use DynaLoader ();
# @ISA = qw(DynaLoader Exporter);

use Exporter();
@ISA = qw(Exporter);
@EXPORT = qw(new handler fixLink fetchURL);


sub new {
  my $class = shift;
  my $self = {};
  bless $self,$class;
  return $self;
}

sub handler {
  my $r = shift;
  
  my $poa = new PAPI::ApachePoA;
  $poa->set_parameters($r);
  $poa->dumpApachePoA("RewritingProxyDump") if ($poa->{'Debug'});
  
  #
  ## Abrimos el fichero de log.
  my $filename = RPLFN.$poa->{'Serv'}.$poa->{'Id'};
  unless (open (RPLF,">$filename")) { 
  	$poa->warning("The log file <$filename> for RewitingProxy cannot be created");
	}
	
  
  #
  # Changes to port PAPI to Apache2 and mod_perl2.
  #
  ## With Apache2 and mod_perl2 the next idiom is proposed:
  ##
  ## my $curl = $r->construct_url;
  ## my $parsedURI = APR::URI->parse($r->pool, $curl); 
  ##
  #my urlToFetch = $r->uri;
  #my $puri = Apache::URI->parse($r);
  #
  my $urlToFetch = $r->uri;
  my $curl = $r->construct_url;
  my $puri = APR::URI->parse($r->pool, $curl);  
  #
  ## End of the changes
  #
   
  my $sch = $puri->scheme;
  for my $urlr (@{$poa->{RejUrlPatterns}}) {
    return FORBIDDEN if ($urlToFetch =~ /$urlr/);
  }
  if (defined $poa->{'Remote_Serv'}) {
    $host_dest = $poa->{'Remote_Serv'};
    my @thn = ($host_dest=~m(^(.*?)://(.*?)\.(.*?)$));
    $name_dest = $thn[1];
  }
  elsif (defined $poa->{'Remote_Dom'}) {
    # Avoid double slashes into URL (seems to have problems in Windows boxes)
    my $rurl = $urlToFetch;
    $rurl =~ s%/(/+)%/%gg;
    $urlToFetch = $rurl;
    
    my @thn = ($urlToFetch =~ m(([^/]+)/(.*?)$));
    $host_dest = $sch . "://" . $thn[0] . "." . $poa->{'Remote_Dom'};
    $name_dest = $thn[0];
    $urlToFetch = "/" . $thn[1];
  }

# $papx_fulldest is meant to be useful for proxy definitions that can be used
# in VirtualHost and/or Location contexts.
  $papx_location = ( $papx_pathdest eq "/" ) ? "" : $papx_pathdest;
  $papx_fulldest = "$papx_location/$papx_namedest";
  $papx_fulldest =~ s%/(/+)%/%gg;
# Keep compatibility with older proxy specs
  $host_dest = $papx_hostdest;
  $name_dest = $papx_namedest;
  $path_dest = $papx_pathdest;  

  $self->headerDebug($urlToFetch);
  
  return &fetchURL($r, $poa, $urlToFetch);
}


###################################################
# fetchURL
#
# Parameters:
#
# $r - an Apache request object
# $poa - A PAPI ApachePoa object
# $url - the URL to fetch and process
#
# Returns:
#
# OK if it's happy.
# an appropriate code if other than 200 is received.
sub fetchURL {
  my $r = shift;
  my $poa = shift;
  my $url = shift;
  
# Set local IP address if required by configuration
  if (defined $poa->{PxIP} && $poa->{PxIP} ne "") {
    @LWP::Protocol::http::EXTRA_SOCK_OPTS = (LocalAddr => $poa->{PxIP});
  }
  my $ua = new LWP::UserAgent;

# Let's make sure we send all redirects back to the browser
  $ua->requests_redirectable([]);
# Let's carry the User Agent to the server also.
  
  #
  ## Changes to port PAPI to Apache2 and mod_perl2
  #
  ## A blessed objet is waitted, or a ARP::Table derivate, so there are to pass a
  ## $r->headers_in->{'key'} that is a derivate os an ARP:Table.
  ##
  ## The original code was:
  ##
  ##		  my $browser = ($r->header_in->{User-Agent} || '');
  ##
  my $browser = '';
  if ( defined $r->headers_in->{'User-Agent'} ){
      $browser = $r->headers_in->{'User-Agent'};
  }
  #
  ## End of Changes for Apache 2 and Mod_perl2
  #
  
  $ua->agent($browser);
# Use a proxy server if defined
  if (defined $poa->{Proxy} && $poa->{Proxy} ne "") {
    $ua->proxy ('http', $poa->{Proxy});
  }
# Use HTTP authentication if defined
  my $tauri = URI::URL->new($papx_hostdest);
  #
  ## Changes made by Dani due to XML PAPI configuration mode. 18 July 2006
  #
  
  for my $hauth (@{$poa->{HttpAuth}}) {
    #my ($br,$bu,$bp) = split(/\s*::\s*/,$hauth);
    my ($br,$bu,$bp) = ($hauth->{br}, $hauth->{bu}, $hauth->{bp});
    if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
      my $brv = eval $br; if (!defined $brv) { $brv = $br; }
      my $buv = eval $bu; if (!defined $buv) { $buv = $bu; }
      my $bpv = eval $bp; if (!defined $bpv) { $bpv = $bp; }
			$self->debugRegex("HTTPAuth REALM", $br, $brv, "EvalPx");
			$self->debugRegex("HTTPAuth USER", $bu, $buv, "EvalPx");
			$self->debugRegex("HTTPAuth PASSWORD", $bp, $bpv, "EvalPx");
      $ua->credentials($tauri->host_port(),$brv,$buv,$bpv);
    }
    else { $ua->credentials($tauri->host_port(),$br,$bu,$bp); }
  }
  my $rurl = $url;
# Avoid double slashes into URL (seems to have problems in Windows boxes)
  $rurl =~ s%/(/+)%/%gg;
  my $turl = "";
  if (defined $poa->{StripLocation} && $poa->{StripLocation} eq "1") { $turl = $rurl; }
  else { $turl = "$papx_pathdest$rurl"; }
# Chase again double slashes
  $turl =~ s%/(/+)%/%gg;
  $url="$papx_hostdest$turl";
  $rurl = $url;
# We have to append the query string since it got munged by
# apache when this was first requested.
  $url .= "?". $r->args if ($r->args && $url !~ /\?/);
  my ($up1, $hostname) = split (/\/\//, $papx_hostdest);

  my($res,$req);
  $self->debug("Requested URL by HTTP::Request: <$url>");
  $poa->accesslog("Fetching remote URL: $url"); 
  $req = HTTP::Request->new($r->method, $url);
# Process and tidy headers:
# Do not pass to the remote server Lcook and Hcook
# Rewrite "Host:" to the remote host if present
# Block all "Connection:" headers
# Since we need to parse the HTML response, block "Accept-Encoding", so
# we don't have to uncompress data before parsing it

#
## Changes to port PAPI to Apache2 and mod_perl2
#
#  Since mod_perl 2.0.0 $r->headers_in(); return an APR::Table object, not a
#  HASH. So  The process of the headers must change.
#
#  OLD CODE:   
#    my %hdrs = $r->headers_in();
#    for my $hdr (keys %hdrs) {
  my $hdrs = $r->headers_in();
  for my $hdr (keys %{$hdrs}) {
    if ($hdr eq 'Cookie') {
      my $fcv = "";
#
#  OLD CODE:
#    for my $ckv (split /[\s]*;[\s]*/,$hdrs{$hdr}) {
      for my $ckv (split /[\s]*;[\s]*/,$hdrs->{$hdr}) {
        my ($cn, $cv) = split /[\s]*=[\s]*/,$ckv,2;
        if ($cn !~ /^Hcook$/i && $cn !~ /^Lcook$/i) {
          if ($fcv ne "") { $fcv .= "; $cn=$cv"; }
          else { $fcv = "$cn=$cv"; }
        }
      }
      $req->header ('Cookie', $fcv) if ($fcv ne "");
    }
    elsif ($hdr eq 'Host') {
      my ($phn, $op) = split /\//,$hostname,2;
      $req->header ('Host', $phn);
    }
    elsif ($hdr ne 'Connection' && $hdr ne 'Accept-Encoding') {
      $req->header ($hdr, $hdrs->{$hdr});
    }
  }
#
## End of Changes for Apache 2 and Mod_perl2
#
  if ($r->method eq 'POST'){
	#
  	## Changes to port PAPI to Apache2 and mod_perl2
    	#
  	#my $len = $r->header_in('Content-length');
    my $len = $r->headers_in->{'Content-length'};
    #
    ## End of the changes
    #
    my $buf;
    $r->read($buf,$len);
    $req->content($buf);
  }
  $ua->{PAPICallBackList}=[$r,$poa,$url];
  $acum=undef;
  $res = $ua->request($req,\&chunkCallBack,$poa->{PxCkSize});
  if (($res->header('X-PAPICallBackCtl') ==2) || ($res->header('X-PAPICallBackCtl') ==3)){ return(OK); }
  $res->content($acum) if ($acum);
# Match form processor with URL without query string
  my $formProc = undef;
  foreach my $u (keys %{$poa->{formProc}}) {
    if ( $url =~ $u ) {
      $formProc = $poa->{formProc}{$u};
      $self->debug("Form_Processor mathed: hashKey <$u>\n\thashValue <$formProc>\n\tURL <$url>");
      last;
    }
  }

  if (defined $formProc) {
# Get any cookie received with the form
    my @rsc = ();
    $res->scan(sub {
      my $hdr = shift;
      my $val = shift;
      if ($hdr =~ /Set-Cookie/) {
        my @cva = split /[\s]*;[\s]*/, $val;
        for my $cv (@cva) {
          if ($cv !~ /^domain=/i && $cv !~ /^path=/i && $cv !~ /^expires=/i) {
            push @rsc, $cv;
          }
        }
      }
    });
    $poa->accesslog("Processing form at $url");
    $self->debug("Processing form at $url");
    my $tree = HTML::TreeBuilder->new_from_content($res->content);
    my @forms = $tree->find_by_tag_name('form');
    my $g = $forms[0];
    my $withname = ($formProc ne "")? 1 : 0;
    my $found = 0;
    for my $f (@forms) {
      if ($f->attr('name') eq $formProc->{formName}) {
        $g = $f;
        $found = 1;
        last;
      }
    }
    if (($withname && $found) || !$withname) {
      $poa->{'log'}->debug("The URL for processing the FORM: $url;") if ($poa->{'Debug'});
      $self->->debug("The URL for processing the FORM: $url;");
      my $thef = HTTP::Request::Form->new($g, $url);
      $poa->{'log'}->debug("The FORM base: ".$thef->base()) if($poa->{'Debug'});
      $self->debug("The FORM base: ".$thef->base());
      $poa->{'log'}->debug("The FORM link: ".$thef->link()) if($poa->{'Debug'});
      $self->debug("The FORM link: ".$thef->link());
      my @buttons = $thef->buttons();
      my $button;
# When Form_Processor has a field named button this field value
# is treated as the name of buttons
      my $cont=0;
      if (defined $formProc->{buttonName} || defined $formProc->{buttonIndex}) {
        if (defined $formProc->{buttonName}) {
          $button=$formProc->{buttonName};
        }elsif (defined $formProc->{buttonIndex}) {
          $button=$buttons[$formProc->{buttonIndex}];
        }
        foreach my $fb (@buttons) {
          if ($fb eq $button) { $cont=1; last; }
        }
      }else{
        $cont=1;
        $button=$buttons[0];
      }
      if ($cont) {
        $poa->accesslog("Sending form data to ".$thef->link);
        $self->debug("Sending form data to ".$thef->link);
        for my $f (keys %{$formProc->{settings}}) {
          my $fv = undef;
          if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
            $fv = eval $formProc->{settings}{$f};
            $self->debugRegex("Forms", $f, $fv, "EvalPx");
          }
          if (!defined $fv){
            $fv = $formProc->{settings}{$f};
          }
          $thef->field($f, $fv);
        }
        $poa->accesslog("Pressing the button: $button");
        $self-debug("Pressing the button: $button");
        my $freq = $thef->press($button);    
# Send along with the new request the cookies received from the client
# in the original request and those received from the remote server in the
# initial response

#
## Changes to port PAPI to Apache2 and mod_perl2
#
#  Since mod_perl 2.0.0 $r->headers_in(); return an APR::Table object, not a
#  HASH. So  The process of the headers must change.
#
#  OLD CODE:   
#    for my $hdr (keys %hdrs) {
        for my $hdr (keys %{$hdrs}) {
          if ($hdr eq 'Cookie') {
            my $fcv = "";
# OLD CODE:
#     for my $ckv (split /[\s]*;[\s]*/,$hdrs{$hdr}) {
     
            for my $ckv (split /[\s]*;[\s]*/,$hdrs->{$hdr}) {
              my ($cn, $cv) = split /[\s]*=[\s]*/,$ckv,2;
              if ($cn !~ /Hcook/i && $cn !~ /Lcook/i) {
                if ($fcv ne ""){
                  $fcv .= "; $cn=$cv";
                }
                else{
                  $fcv = "$cn=$cv";
                }
              }
            }           
            $freq->header ('Cookie', $fcv) if ($fcv ne "");
          }
        }
#
## End of Changes for Apache 2 and Mod_perl2
#        
        for my $cv (@rsc){
          $freq->header ('Cookie', $cv);
        }
	    $ua->{PAPICallBackList}=[$r,$poa,$url];
        $acum=undef;
        $res = $ua->request($freq,\&chunkCallBack,$poa->{PxCkSize});
        if (($res->header('X-PAPICallBackCtl') == 2) || ($res->header('X-PAPICallBackCtl') ==3)){
	      return(OK);
        }
	    $res->content($acum) if ($acum);
      }else{
        $poa->error ("Button $button does not exist in this form, sending form data to ".$thef->link." abandoned");
        $self->debug("Button $button does not exist in this form, sending form data to ".$thef->link." abandoned");
      }
    }
  }
  $r->status($res->code);
  procHeaders($res,$poa,$r,$url);
  my $outString = "";
#
# Parse and rewrite HTML content
$self->debug("Parse and rewrite HTML content");
  if ($res->content_type =~ /html/i) {
    my $content = $res->content;
    my $baseHref = "";
    my $inScript = 0;
    my $p = HTML::TokeParser->new(\$content)
            || {$poa->error("No content available from remote URL $url: $!");
                $self->debug("No content available from remote URL $url: $!");}
# Avoid entity decoding inside attributes
    $p->attr_encoded(1);
# Unless No_XML is set to 1, allow XHTML tags (<x/>)
    if ((defined $poa->{NoXML} && $poa->{NoXML} eq "1")) { $p->xml_mode(0); }
    else { $p->xml_mode(1); }
    while (my $tolkens = $p->get_token ) {
      if ($tolkens->[0] eq 'T' || $tolkens->[0] eq 'C') {
        if ($inScript || (defined $poa->{RwAll} && $poa->{RwAll} eq "1")) {
        	$self->debug("Begin redirects processing");
          for my $redirect (@{$poa->{redirects}}) {
            $self->debug("PAPI_Redirect: Regex<".$redirect->{regex}."> Value<".$redirect->{value}.">");
            my $rde = ""; my $rdv = "";
            if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
              $rde = eval $redirect->{regex};
              $self->debugRegex("PAPI_Redirect regex", $redirect->{regex}, $rde, "EvalPx");
              if (!defined $rde) { $rde = $redirect->{regex}; }
              $rdv = eval $redirect->{value};
              $self->debugRegex("PAPI_Redirect value", $redirect->{value}, $rdv, "EvalPx");
              if (!defined $rdv) { $rdv = $redirect->{value}; }
            }
            else {
              $rde = $redirect->{regex};
              $rdv = $redirect->{value};
            }
            my $evs = "\$tolkens->[1] =~ s($rde)($rdv)gsx";
            eval $evs;
            $self->debugRegex("PAPI_Redirect apply", $tolkens->[1], $evs, "s($rde)($rdv)gsx");
            $self->debug("PAPI_Redirect: finish this one");
          }
        }
        $outString .= $tolkens->[1];
      }
      elsif ($tolkens->[0] eq 'E') {
# End script rewriting
        if ($tolkens->[1] =~ /^script$/i ||
            $tolkens->[1] =~ /^style$/i) { $inScript = 0; }
# Only write tag out if we are in a real end token (XML mode forces this check)
        $outString .= "</" . $tolkens->[1] . ">" if ($tolkens->[2] ne "");
      }
      elsif ($tolkens->[0] eq 'S') {
# Convert all attribute names to lowercase, so the references inside work
# with XML mode on
        for my $ki (keys %{$tolkens->[2]}) {
          my $kj = lc $ki;
          if ($ki ne $kj) {
            $tolkens->[2]{$kj} = $tolkens->[2]{$ki};
            delete $tolkens->[2]{$ki};
          }
        }
# End script rewriting. It will be activated if starting a script tag
        $inScript = 0;
        my $text = $tolkens->[4];
        my @oldLink = ();
        my @newLink = ();
        if ($tolkens->[1] =~ /^a$/i) {
          if ($tolkens->[2]{href}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{href},$url);
            push @oldLink, regexEscape($tolkens->[2]{href});
          }
          if ($tolkens->[2]{onclick}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{onclick},$url);
            push @oldLink, regexEscape($tolkens->[2]{onclick});
          }
        }
        elsif ($tolkens->[1] =~ /^base$/i) {
          if ($tolkens->[2]{href}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{href},$url);
            push @oldLink, regexEscape($tolkens->[2]{href});
            $url = $tolkens->[2]{href};
          }
        }
        elsif ($tolkens->[1] =~ /^meta$/i) {
          if ($tolkens->[2]{content} =~ /url/i) {
            my ($junk,$tmpLink) = split (/=/, $tolkens->[2]{content});
            push @newLink, &fixLink($r,$poa,$tmpLink,$url);
            push @oldLink, regexEscape($tmpLink);
          }
        }
        elsif ($tolkens->[1] =~ /^(area|link)$/i) {
          if ($tolkens->[2]{href}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{href},$url);
            push @oldLink, regexEscape($tolkens->[2]{href}); 
          }
        }
        elsif ($tolkens->[1] =~ /^(frame|img|iframe)$/i) {
          if ($tolkens->[2]{src}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{src},$url);
            push @oldLink, regexEscape($tolkens->[2]{src});
          }
        }
        elsif ($tolkens->[1] =~ /^form$/i) {
          if ($tolkens->[2]{action}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{action},$url);
            push @oldLink, regexEscape($tolkens->[2]{action});
          }
        }
        elsif ($tolkens->[1] =~ /^(table|td|body)$/i) {
          if ($tolkens->[2]{background}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{background},$url);
            push @oldLink, regexEscape($tolkens->[2]{background});
          }
        }
# Activate rewriting for the content of style specs (imported URLs...)
        elsif ($tolkens->[1] =~ /^style$/i) { $inScript = 1; }
        elsif ($tolkens->[1] =~ /^script$/i) {
          if ($tolkens->[2]{src}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{src},$url);
            push @oldLink, regexEscape($tolkens->[2]{src});
          }
# Inline script (not from a source object). Activate rewriting
          else { $inScript = 1; }
        }
        elsif ($tolkens->[1] =~ /^option$/i) {
          if ($tolkens->[2]{value}) {
            push @newLink, &fixLink($r,$poa,$tolkens->[2]{value},$url);
            push @oldLink, regexEscape($tolkens->[2]{value});
          }
        }
        elsif ($tolkens->[1] =~ /^input$/i) {
          if ($tolkens->[2]{value}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{value},$url);
            push @oldLink, regexEscape($tolkens->[2]{value});
          }
          if ($tolkens->[2]{src}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{src},$url);
            push @oldLink, regexEscape($tolkens->[2]{src});
          }
        }
        elsif ($tolkens->[1] =~ /^object$/i) {
          if ($tolkens->[2]{archive}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{archive},$url);
            push @oldLink, regexEscape($tolkens->[2]{archive});
          }
          if ($tolkens->[2]{codebase}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{codebase},$url);
            push @oldLink, regexEscape($tolkens->[2]{codebase});
          }
          if ($tolkens->[2]{data}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{data},$url);
            push @oldLink, regexEscape($tolkens->[2]{data});
          }
          if ($tolkens->[2]{map}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{map},$url);
            push @oldLink, regexEscape($tolkens->[2]{map});
          }
        }
        elsif ($tolkens->[1] =~ /^param$/i) {
          if ($tolkens->[2]{value}) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{value},$url);
            push @oldLink, regexEscape($tolkens->[2]{value});
          }
        }                
# Process script attributes
        for my $att (@{$tolkens->[3]}) {
          if ($att =~ /^on/i) {
            push @newLink, &fixAnchor($r,$poa,$tolkens->[2]{$att},$url);
            push @oldLink, regexEscape($tolkens->[2]{$att});
          }
        }
# Some servers send character entities inside URLs, that make the regex
# substiting $oldLink by $newLink fail. Up to now, only '&amp;' -> '&'
# is considered.
# Not longer needed in the current parser configuration
#        if ($#newLink >= 0) { $text =~ s(\&amp;)(&)gs; }
          for my $i (0..$#newLink) { $text =~ s($oldLink[$i])($newLink[$i])gs; }
          $outString .= $text;
        }
        else { $outString .= $tolkens->[4]; }
    }
  }
#
# Non-HTML content. Check for MIME types and URL patterns
# where rewriting is applicable
  else {
    $outString = $res->content;
    for my $redirect (@{$poa->{redirects}}) {
      my $rde = ""; my $rdv = "";
      if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
        $rde = eval $redirect->{regex};
        if (!defined $rde) { $rde = $redirect->{regex}; }
        $rdv = eval $redirect->{value};
        if (!defined $rdv) { $rdv = $redirect->{value}; }
      }
      else {
        $rde = $redirect->{regex};
        $rdv = $redirect->{value};
      }
      my $evs = "\$outString =~ s($rde)($rdv)gsx";
      eval $evs;
    }
  }
  #
  ## Changes to port PAPI to Apache2 and mod_perl2
  #
  #$r->header_out('Content-length'=> length($outString));
  $r->headers_out->set('Content-length'=> length($outString));
  #
  ## End of the changes
  #
# Some browsers are reported to be extremely sensitive to content types
# containing extra blank characters. Clean it before sending
  my $ttc = $res->content_type;
  $ttc =~ s/(.*?)[\s]+$/$1/;
  $r->content_type($ttc);
  #
  ## Changes to port PAPI to Apache2 and mod_perl2
  #
  ## Apache2 and mod_perl2 make this unnecessary
  ## The headers must not be explicitly sent. The header only need to be defined with
  ## the method $r->content_type($type).
  #
  #$r->send_http_header;
  #
  ## Removed method
  
  #
  ## End of the changes
  #
  
  #
  ## Cerramos el fichero de log
	close(RPLF);
  
  $r->print($outString);
  return(OK);
}


#################################
# chunkCallBack
# Manage chunked transfer of data
sub chunkCallBack {
  my ($data,$response,$protocol) = @_;
  my $cont=$response->header('X-PAPICallBackCtl');
  my ($r,$poa,$url)=@{$protocol->{ua}->{PAPICallBackList}};
  if(!$cont){
    $cont=2;
    if ($response->header('content_type') =~ /html/i){
      $response->header('X-PAPICallBackCtl',1);
      $cont=1;
    }
    elsif ($response->content_type =~ /$poa->{RewMimeTypes}/i) {
      $response->header('X-PAPICallBackCtl',1);
      $cont = 1; 
    }
    else{
      for my $patt (@{$poa->{RewUrlPatterns}}) {
        if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
          my $rpt = eval $patt;
          if (!defined $rpt) { $rpt = $patt; }
          if ($url =~ /$rpt/) { $cont = 1; last; }
        }
        else { if ($url =~ /$patt/) { $cont = 1; last; } }
      }
    }
    $response->header('X-PAPICallBackCtl',$cont);
    if ($cont == 2){
      $r->status($response->code);
      procHeaders($response,$poa,$r,$url);
      $r->content_type($response->content_type);
      #
      ## Changes to port PAPI to Apache2 and mod_perl2
      #
      # $r->send_http_header;
      # Removed method
      #
      ## End of the changes
      #
    
    }
  }
  if($cont == 1){ $acum .= $data; }
  else{
    my $c=$r->connection;
    if($c->aborted){
      $response->header('X-PAPICallBackCtl',3);
      $poa->error("Chunked transfer aborted"); 
      die("Chunked transfer aborted. Connection closed");
    }
    $r->print($data);
  }
}


############################################################################
# procHeaders
# Process and tidy received headers:
# Cookies are transformed so they can be sent back to the originating server
# Redirects (and the non-standard Refresh:) are adapted to the rewriten URLs
sub procHeaders 
{
  my $res = shift;
  my $poa = shift;
  my $r = shift;
  my $url = shift;

  $res->scan(sub {
    my $hdr = shift;
    my $val = shift;
    if ($hdr =~ /Set-Cookie/i) {
      my @cva = split /[\s]*;[\s]*/, $val;
      my $cfv = undef;
      for my $cv (@cva) {
        if ($cv !~ /^domain=/i) {
          if ($cv =~ /^path=/i) {
            if (defined $poa->{'Remote_Dom'}) {
              my ($pc,$vc) = split /[\s]*=[\s]*/, $cv;
              $cv = "path=/" . $name_dest . $vc;
            }
          }
          if (!defined $cfv){ $cfv = $cv; }
          else { $cfv .= "; $cv"; }
        }
      }
      $val = $cfv;
      #
      ## Changes to port PAPI to Apache2 and mod_perl2
      #
      # OLD CODE:
      #     $r->cgi_headers_out ($hdr, $val);
      #
      $r->err_headers_out->set($hdr => $val);
      # Cahnged method
      #
      ## End of the changes
      #
    }
    elsif ($hdr =~ /Location/i) {
      $val = &fixLink($r,$poa,$val,$url);
      #
      ## Changes to port PAPI to Apache2 and mod_perl2
      #
      #$r->header_out ($hdr, $val);
      $r->headers_out->set($hdr => $val);
      # Cahnged method
      #
      ## End of the changes
      #
    }
    elsif ($hdr =~ /Refresh/i) {
      my @rva = split /[\s]*;[\s]*/,$val;
      my ($ui,@rv) = split /[\s]*=[\s]*/,$rva[1];
      my $rvv = join "=",@rv;
      $rva[1] = "URL=".&fixLink($r,$poa,$rvv,$url);
      $val = join ";",@rva;
      #
      ## Changes to port PAPI to Apache2 and mod_perl2
      #
      #$r->header_out ($hdr, $val);
      $r->headers_out->set($hdr => $val);
      # Cahnged method
      #
      ## End of the changes
      #
    }
    #
    ## Changes to port PAPI to Apache2 and mod_perl2
    #
    #else { $r->header_out ($hdr, $val);
    else{
      $r->headers_out->set($hdr => $val);
    }
    # Cahnged method
    #
    ## End of the changes
    #
  });
}


##################################################################
# fixAnchor
# This is a wrapper to fixLink for dealing with JavaScript anchors
sub fixAnchor {
  my $r = shift;
  my $poa = shift;
  my $link = shift;
  my $baseLink = shift;

  my $rtv = &fixLink($r, $poa, $link, $baseLink);
# If a special method is in use, apply redirects
# At the moment, only "javascript:" is considered
  if ($rtv =~ /^javascript:/i) {
    for my $redirect (@{$poa->{redirects}}) {
      my $rde = ""; my $rdv = "";
      if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
        $rde = eval $redirect->{regex};
        if (!defined $rde) { $rde = $redirect->{regex}; }
        $rdv = eval $redirect->{value};
        if (!defined $rdv) { $rdv = $redirect->{value}; }
      }
      else {
        $rde = $redirect->{regex};
        $rdv = $redirect->{value};
      }
      my $evs = "\$rtv =~ s($rde)($rdv)gsx";
      eval $evs;
    }
  }
  return $rtv;
}


###############################
# sub fixLink
#
# Parameters:
#
# $r - an Apache request object.
# $poa - a PAPI ApachePoA object
# $url - a reference to another resource.
#     It can be a URL or a relative reference.
#     fixLink will do the right thing with it.
# $baseLink - the link pointing to the current resource.
#
# Returns:
# a fixed URL using this module again.
#################################
sub fixLink {
  my $r = shift;
  my $poa = shift;
  my $link = shift;
  my $baseLink = shift;
  my $server_name = $r->get_server_name;
  my $i;
  my $urlPath = "";
  my $hostName = "";
  my $base_name=$r->location();
  my $tdn = undef;
  
  #
  ## Modificacion para portar a Apache2
  ## Se ha de construir el URL a parsear....
  #
  
  # Si obnenemos un objeto Server a partir de $r->server() el puerto que obtenemos
  # no es el correcto, sino que devuelve 0; por lo que el objeto server lo
  # obtenemos por medo se la siguiente linea.
  my $serv = Apache2::ServerUtil->server;
  
  #my $puri = Apache::URI->parse($r);
  
  my $curl = $r->construct_url;
  my $puri = APR::URI->parse($r->pool, $curl);
  #
  ##

  my $sch = $puri->scheme;
  my ($protocol, $hostname, @dirs);
  
  ## Clean dobule slash //
 	$link =~ s/([^:])\/+/$1\//g;
 	$baseLink =~ s/([^:])\/+/$1\//g;
 	$curl =~ s/([^:])\/+/$1\//g;
  
  ($protocol, $hostname) = split (/\/\//, $baseLink);
  ($hostname,@dirs) = split (/\//, $hostname);
 
  
  if ($dirs[1] =~ /\w/) { $urlPath = join ("/", @dirs); }
  else { $urlPath = $dirs[0] unless ($baseLink =~ /\w$/); }
  if ($baseLink =~ /\w$/ && $dirs[1] =~ /\w/) {
    for ($i = length($urlPath); $i > 0; $i--) {
      last if (substr($urlPath, $i, 1) eq "/");
    }
    $urlPath = substr ($urlPath, 0, $i) if ($i && $urlPath =~ /\//);
  }

  if (defined $poa->{'Remote_Dom'}) {
    $tdn = $poa->{'Remote_Dom'};
    my ($thn) = ($hostname=~m(^(.+?)\.($tdn)$));
    $base_name = "/".$thn.$base_name;
  }

  #Apache2 makes this unnecessary
  #$base_name=~ s/$path_dest$//;
  
  $hostname = "$protocol//$hostname";
  if (($sch eq 'http' && $serv->port != 80) || ($sch eq 'https' && $serv->port != 443)){ 
    $server_name .= ":".$serv->port();
  }
  
  if ($link =~ /^\//) {
    $link = $sch."://".$server_name.$base_name.$link;
  }
  elsif ($link =~ /^\.\./) {
    $link = $sch."://".$server_name.$base_name."/".$urlPath."/".$link;
  }
  elsif ($link =~ /^$host_dest/i) {
    $link =~ s/^$host_dest//i;
    $link = $sch."://".$server_name.$base_name.$link;
  }
  elsif (defined $tdn && $link =~ /^\w+?\.$tdn/i) {
    my ($up1, $up2) = split (/\/\//, $link);
    my ($thn) = ($up2=~m(^(.+?)\.($tdn)));
    $link =~ s/^(.+?)\.$tdn//i;
    $link = $sch . "://" . $server_name . "/" . $thn . $link;
  }
  
  for my $redirect (@{$poa->{redirects}}) {
    my $rde = ""; my $rdv = "";
    if ((defined $poa->{EvalPx} && $poa->{EvalPx} eq "1")) {
      $rde = eval $redirect->{regex};
      if (!defined $rde) { $rde = $redirect->{regex}; }
      $rdv = eval $redirect->{value};
      if (!defined $rdv) { $rdv = $redirect->{value}; }
    }
    else {
      $rde = $redirect->{regex};
      $rdv = $redirect->{value};
    }
    my $evs = "\$link =~ s($rde)($rdv)gsx";
    eval $evs;
  }
  ## Clean dobule slash //
 	$link =~ s/([^:])\/+/$1\//g;
  return $link;
}


############################################################
# regexEscape
# This silly little regex fixes \-[]()*+?|&# in the URL.
# Without this regex, any of these characters in a URL will
# cause a server error (unless they resolve into something
# sensible to regex, in which case the server does something
# magical and unpredictable with the URL)
sub regexEscape {
  my $url = shift;
  $url =~ s/(\\|\-|\[|\]|\(|\)|\*|\+|\?|\||\&|\#)/\\$1/g;
  return $url;
}

############################################################
# debugRegex
# This function sends a log message to a file for debug purpouse
# what: What have been rewite
# original: The Original data to be rewite
# rewrote: The resultant data of rewriting
# regex: The REGEX used to trewrite the data
#
# 

sub debugRegex {
  my $what = shift;
  my $original = shift;
  my $rewrote = shift;
  my $regex = shift;

  my $outString = '';
  my $currentTime = '';

	
	my $currentTime = gmtime();
	
	$outStrign = "[".$currentTime."] ".$msg."\n";
	print RPLF, $outString;
}

############################################################
# debugRegex
# This function sends a log message to a file for debug purpouse
#
# mesg: The mesage to be logged
sub debug {
	my $msg = shift;
	
	my $currentTime = gmtime();
	
	$outStrign = "[".$currentTime."] ".$msg."\n";
	print RPLF, $outString;

}


############################################################
# headerDebug
# This function make the header for the debug file
sub headerDebug {
	my $localURL = shift;
	
	my $outString = "############################################################\n";
	$outString   .= "#                                                           \n";
	$outString   .= "# URL to fetch:                                             \n";
	$outString   .= "#     $localURL\n";
	$outString   .= "#                                                           \n";
	$outString   .= "############################################################\n";
	
	print RPLF, $outString;
	
}

1;

__END__
